/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.lang.schema.internal;

import com.redhat.thermostat.lang.schema.SchemaBuilder;
import com.redhat.thermostat.lang.schema.SchemaValidationException;

import com.redhat.thermostat.lang.schema.annotations.Maximum;
import com.redhat.thermostat.lang.schema.annotations.Minimum;
import com.redhat.thermostat.lang.schema.annotations.Schema;
import com.redhat.thermostat.lang.schema.annotations.Type;
import com.redhat.thermostat.lang.schema.internal.SchemaBuilderImpl;
import com.redhat.thermostat.lang.schema.models.Timestamp;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import static org.junit.Assert.assertEquals;

/**
 */
public class SchemaBuilderTest {

    @Type(description = "Some Object for testing out")
    public class SomeMoreComplexType {
        @Schema(description = "Some Long", required = true)
        private long aLong;

        @Schema(description = "Some String", required = true)
        private String aString;

        public SomeMoreComplexType(long aLong, String aString) {
            this.aLong = aLong;
            this.aString = aString;

        }

        public long getALong() {
            return aLong;
        }

        public String getAString() {
            return aString;
        }
    }

    @Type(description = "Some Object for testing out")
    public class SomeMoreComplexType2 {
        @Schema(description = "Some Long", required = true)
        private long aLong;

        // not part of the schema, won't be serialised!
        private String aString;

        public SomeMoreComplexType2(long aLong, String aString) {
            this.aLong = aLong;
            this.aString = aString;

        }

        public long getALong() {
            return aLong;
        }

        public String getAString() {
            return aString;
        }
    }

    @Type(description = "Something more complex perhaps?")
    public class CompoundType {
        @Schema(description = "Some Long", required = true)
        private long aLong;

        @Schema
        private SomeMoreComplexType complexType;

        public CompoundType(long aLong, SomeMoreComplexType complexType) {
            this.aLong = aLong;
            this.complexType = complexType;

        }

        public long getALong() {
            return aLong;
        }

        public SomeMoreComplexType getComplexType() {
            return complexType;
        }
    }

    @Type(description = "Something more complex perhaps?")
    public class CompoundTypeMissingSchema {
        @Schema(description = "Some Long", required = true)
        private long aLong;

        // won't be serialised
        private SomeMoreComplexType complexType;

        public CompoundTypeMissingSchema(long aLong, SomeMoreComplexType complexType) {
            this.aLong = aLong;
            this.complexType = complexType;
        }

        public long getALong() {
            return aLong;
        }

        public SomeMoreComplexType getComplexType() {
            return complexType;
        }
    }

    @Type(description = "An extended type that should inherit its schema")
    public class ExtendedType extends CompoundType {

        @Schema(description = "Some other long from derived class", required = true)
        private long anotherLong;

        public ExtendedType(long aLong, long anotherLong, SomeMoreComplexType complexType) {
            super(aLong, complexType);
            this.anotherLong = anotherLong;
        }
    }

    private SchemaBuilder service;

    @Before
    public void setUp() {
        service = new SchemaBuilderImpl(false);
    }

    @Test(expected = java.lang.IllegalArgumentException.class)
    public void testNoTypeAnnotation() {
        String test = "primitive type";
        service.getSchema(test.getClass());
    }

    @Test
    public void testSimpleTypeSchema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"Timestamp\",\"description\":\"A Type representing the number of milliseconds elapsed since the Unix Epoch\",\"type\":\"object\",\"properties\":{\"timestamp\":{\"description\":\"The number of milliseconds elapsed since the Unix Epoch\",\"type\":\"long\",\"minimum\":0}},\"required\":[\"timestamp\"]}",
                service.getSchema(Timestamp.class));
    }

    @Test
    public void testComplexTypeSchema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"SomeMoreComplexType\",\"description\":\"Some Object for testing out\",\"type\":\"object\",\"properties\":{\"aLong\":{\"description\":\"Some Long\",\"type\":\"long\"},\"aString\":{\"description\":\"Some String\",\"type\":\"String\"}},\"required\":[\"aLong\",\"aString\"]}",
                service.getSchema(SomeMoreComplexType.class));
    }


    @Test
    public void testComplexType2Schema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"SomeMoreComplexType2\",\"description\":\"Some Object for testing out\",\"type\":\"object\",\"properties\":{\"aLong\":{\"description\":\"Some Long\",\"type\":\"long\"}},\"required\":[\"aLong\"]}",
                service.getSchema(SomeMoreComplexType2.class));
    }

    @Test
    public void testCompoundTypeSchema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"CompoundType\",\"description\":\"Something more complex perhaps?\",\"type\":\"object\",\"properties\":{\"aLong\":{\"description\":\"Some Long\",\"type\":\"long\"},\"complexType\":{\"description\":\"\",\"type\":\"SomeMoreComplexType\"}},\"required\":[\"aLong\"]}",
                service.getSchema(CompoundType.class));
    }

    @Test
    public void testCompoundTypeMissingSchemaSchema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"CompoundTypeMissingSchema\",\"description\":\"Something more complex perhaps?\",\"type\":\"object\",\"properties\":{\"aLong\":{\"description\":\"Some Long\",\"type\":\"long\"}},\"required\":[\"aLong\"]}",
                service.getSchema(CompoundTypeMissingSchema.class));
    }


    @Test
    public void testExtendedTypeSchema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"ExtendedType\",\"description\":\"An extended type that should inherit its schema\",\"type\":\"object\",\"extends\":\"CompoundType\",\"properties\":{\"anotherLong\":{\"description\":\"Some other long from derived class\",\"type\":\"long\"}},\"required\":[\"anotherLong\"]}",
                service.getSchema(ExtendedType.class));
    }

    @Type(description = "Some Type for testing out")
    public class ATypeWithConstraints {
        @Schema(description = "Some Long", required = true)
        @Minimum(2)
        @Maximum(value = 5, exclusive = true)
        private long aLong;

        public ATypeWithConstraints(long aLong) {
            this.aLong = aLong;
        }

        public long getALong() {
            return aLong;
        }
    }

    @Test
    public void testTypeWithConstraintsSchema() {
        assertEquals("{\"$schema\":\"http://json-schema.org/draft-04/schema#\",\"title\":\"ATypeWithConstraints\",\"description\":\"Some Type for testing out\",\"type\":\"object\",\"properties\":{\"aLong\":{\"description\":\"Some Long\",\"type\":\"long\",\"minimum\":2,\"maximum\":5,\"exclusiveMaximum\":true}},\"required\":[\"aLong\"]}",
                service.getSchema(ATypeWithConstraints.class));
    }

    @Type(description = "Some Type for testing out")
    public class ATypeWithBadConstraints {
        @Schema(description = "Some Long", required = true)
        @Minimum(5)
        @Maximum(value = 1, exclusive = true)
        private long aLong;

        public ATypeWithBadConstraints(long aLong) {
            this.aLong = aLong;
        }

        public long getALong() {
            return aLong;
        }
    }

    @Test(expected = SchemaValidationException.class)
    public void testTypeWithBadConstraintsSchema() {
        service.getSchema(ATypeWithBadConstraints.class);
    }
}
